@@ -1,3 +1,5 @@
+from __future__ import print_function, absolute_import, division
+
 import numpy as np
 
 from .cube_utils import iterator_strategy
@@ -43,14 +45,13 @@ def _slice0(cube, axis):
     result = np.zeros(shp)
 
     view = [slice(None)] * 3
-    pix_size = cube._pix_size()[axis]
 
     valid = np.zeros(shp, dtype=np.bool)
     for i in range(cube.shape[axis]):
         view[axis] = i
         plane = cube._get_filled_data(fill=np.nan, view=view)
         valid |= np.isfinite(plane)
-        result += np.nan_to_num(plane) * pix_size[view]
+        result += np.nan_to_num(plane) * cube._pix_size_slice(axis)
     result[~valid] = np.nan
     return result
 
@@ -72,7 +73,7 @@ def _slice1(cube, axis):
     result = np.zeros(shp)
 
     view = [slice(None)] * 3
-    pix_size = cube._pix_size()[axis]
+    pix_size = cube._pix_size_slice(axis)
     pix_cen = cube._pix_cen()[axis]
     weights = np.zeros(shp)
 
@@ -81,8 +82,8 @@ def _slice1(cube, axis):
         plane = cube._get_filled_data(fill=0, view=view)
         result += (plane *
                    pix_cen[view] *
-                   pix_size[view])
-        weights += plane * pix_size[view]
+                   pix_size)
+        weights += plane * pix_size
     return result / weights
 
 
@@ -99,7 +100,7 @@ def moment_slicewise(cube, order, axis):
     result = np.zeros(shp)
 
     view = [slice(None)] * 3
-    pix_size = cube._pix_size()[axis]
+    pix_size = cube._pix_size_slice(axis)
     pix_cen = cube._pix_cen()[axis]
     weights = np.zeros(shp)
 
@@ -112,8 +113,8 @@ def moment_slicewise(cube, order, axis):
         plane = cube._get_filled_data(fill=0, view=view)
         result += (plane *
                    (pix_cen[view] - mom1) ** order *
-                   pix_size[view])
-        weights += plane * pix_size[view]
+                   pix_size)
+        weights += plane * pix_size
 
     return (result / weights)
 
@@ -126,7 +127,7 @@ def moment_raywise(cube, order, axis):
     out = np.zeros(shp) * np.nan
 
     pix_cen = cube._pix_cen()[axis]
-    pix_size = cube._pix_size()[axis]
+    pix_size = cube._pix_size_slice(axis)
 
     for x, y, slc in cube._iter_rays(axis):
         # the intensity, i.e. the weights
@@ -135,7 +136,7 @@ def moment_raywise(cube, order, axis):
         if not include.any():
             continue
 
-        data = cube.flattened(slc).value * pix_size[slc][include]
+        data = cube.flattened(slc).value * pix_size
 
         if order == 0:
             out[x, y] = data.sum()
@@ -158,7 +159,7 @@ def moment_cubewise(cube, order, axis):
     """
 
     pix_cen = cube._pix_cen()[axis]
-    data = cube._get_filled_data() * cube._pix_size()[axis]
+    data = cube._get_filled_data() * cube._pix_size_slice(axis)
 
     if order == 0:
         return allbadtonan(np.nansum)(data, axis=axis)
