@@ -1,3 +1,5 @@
+from __future__ import print_function, absolute_import, division
+
 import numpy as np
 from astropy.wcs import (WCSSUB_SPECTRAL, WCSSUB_LONGITUDE, WCSSUB_LATITUDE)
 from . import wcs_utils
@@ -191,8 +193,13 @@ class SliceIndexer(object):
         raise Exception("You need to specify a slice (e.g. ``[:]`` or "
                         "``[0,:,:]`` in order to access this property.")
 
-def is_huge(cube, threshold=1e8):
-    if cube.size < threshold:  # smallish
+
+# TODO: make this into a proper configuration item
+# TODO: make threshold depend on memory?
+MEMORY_THRESHOLD=1e8
+
+def is_huge(cube):
+    if cube.size < MEMORY_THRESHOLD:  # smallish
         return False
     else:
         return True
