@@ -1,3 +1,5 @@
+from __future__ import print_function, absolute_import, division
+
 import numpy as np
 import struct
 import warnings
@@ -254,7 +256,7 @@ def read_lmv_tofits(filename):
              fits.header.Card(''.join(s for s in k if s in string.printable),
                               ''.join(s for s in v if s in string.printable)
                               if isinstance(v, six.string_types) else v)
-             for k,v in header.iteritems()
+             for k,v in six.iteritems(header)
              if k not in bad_kws]
     Header = fits.Header(cards)
     hdu = fits.PrimaryHDU(data=data, header=Header)
