@@ -1,3 +1,5 @@
+from __future__ import print_function, absolute_import, division
+
 import warnings
 
 from astropy.io import fits
@@ -8,10 +10,17 @@ from astropy.io.fits.hdu.hdulist import fitsopen as fits_open
 
 import numpy as np
 import datetime
-from .. import version
+try:
+    from .. import version
+    SPECTRAL_CUBE_VERSION = version.version
+except ImportError:
+    # We might be running py.test on a clean checkout
+    SPECTRAL_CUBE_VERSION = 'dev'
+
 from .. import SpectralCube, StokesSpectralCube, LazyMask
 from .. import cube_utils
 
+
 def first(iterable):
     return next(iter(iterable))
 
@@ -26,8 +35,8 @@ def is_fits(input, **kwargs):
     """
     if isinstance(input, six.string_types):
         if input.lower().endswith(('.fits', '.fits.gz',
-                                      '.fit', '.fit.gz',
-                                      '.fits.Z', '.fit.Z')):
+                                   '.fit', '.fit.gz',
+                                   '.fits.Z', '.fit.Z')):
             return True
     elif isinstance(input, (fits.HDUList, fits.PrimaryHDU, fits.ImageHDU)):
         return True
@@ -78,7 +87,7 @@ def read_data_fits(input, hdu=None, **kwargs):
         elif len(arrays) == 1:
             array_hdu = arrays[first(arrays)]
         else:
-            raise ValueError("No table found")
+            raise ValueError("No arrays found")
 
     elif isinstance(input, (fits.PrimaryHDU, fits.ImageHDU)):
 
@@ -96,7 +105,7 @@ def read_data_fits(input, hdu=None, **kwargs):
     return array_hdu.data, array_hdu.header
 
 
-def load_fits_cube(input, hdu=0, meta={}, **kwargs):
+def load_fits_cube(input, hdu=0, meta=None, **kwargs):
     """
     Read in a cube from a FITS file using astropy.
 
@@ -112,6 +121,9 @@ def load_fits_cube(input, hdu=0, meta={}, **kwargs):
 
     data, header = read_data_fits(input, hdu=hdu, **kwargs)
 
+    if meta is None:
+        meta = {}
+
     if 'BUNIT' in header:
         meta['BUNIT'] = header['BUNIT']
 
@@ -130,12 +142,15 @@ def load_fits_cube(input, hdu=0, meta={}, **kwargs):
 
         data, wcs = cube_utils._split_stokes(data, wcs)
 
-        mask = {}
+        stokes_data = {}
         for component in data:
-            data[component], wcs_slice = cube_utils._orient(data[component], wcs)
-            mask[component] = LazyMask(np.isfinite, data=data[component], wcs=wcs_slice)
+            comp_data, comp_wcs = cube_utils._orient(data[component], wcs)
+            comp_mask = LazyMask(np.isfinite, data=comp_data, wcs=comp_wcs)
+            stokes_data[component] = SpectralCube(comp_data, wcs=comp_wcs,
+                                                  mask=comp_mask, meta=meta,
+                                                  header=header)
 
-        cube = StokesSpectralCube(data, wcs_slice, mask, meta=meta, header=header)
+        cube = StokesSpectralCube(stokes_data)
 
     else:
 
@@ -155,8 +170,8 @@ def write_fits_cube(filename, cube, overwrite=False,
         now = datetime.datetime.strftime(datetime.datetime.now(),
                                          "%Y/%m/%d-%H:%M:%S")
         hdu.header.add_history("Written by spectral_cube v{version} on "
-                               "{date}".format(version=version.version,
-                               date=now))
+                               "{date}".format(version=SPECTRAL_CUBE_VERSION,
+                                               date=now))
         hdu.writeto(filename, clobber=overwrite)
     else:
         raise NotImplementedError()
