@@ -2,13 +2,18 @@
 A class to represent a 3-d position-position-velocity spectral cube.
 """
 
+from __future__ import print_function, absolute_import, division
+
 import warnings
 from functools import wraps
 import operator
+import sys
+import re
 
 from astropy import units as u
 from astropy.extern import six
 from astropy.io.fits import PrimaryHDU, ImageHDU, Header, Card, HDUList
+from astropy.utils.console import ProgressBar
 from astropy import log
 from astropy import wcs
 
@@ -57,8 +62,8 @@ def cached(func):
     def wrapper(self, *args):
         # The cache lives in the instance so that it gets garbage collected
         if func not in self._cache:
-            self._cache[func] = func(self, *args)
-        return self._cache[func]
+            self._cache[func, args] = func(self, *args)
+        return self._cache[func, args]
 
     return wrapper
 
@@ -111,39 +116,15 @@ np2wcs = {2: 0, 1: 1, 0: 2}
 class SpectralCube(object):
 
     def __init__(self, data, wcs, mask=None, meta=None, fill_value=np.nan,
-                 header=None, allow_huge_operations=False):
+                 header=None, allow_huge_operations=False, read_beam=True):
 
         # Deal with metadata first because it can affect data reading
         self._meta = meta or {}
-        if 'BUNIT' in self._meta:
 
-            # special case: CASA (sometimes) makes non-FITS-compliant jy/beam headers
-            if self._meta['BUNIT'].lower() == 'jy/beam':
-                self._unit = u.Jy
-                try:
-                    import radio_beam
-                    self.beam = radio_beam.Beam.from_fits_header(header)
-                    self._meta['beam'] = self.beam
-                    warnings.warn("Units were JY/BEAM.  The 'beam' is now "
-                                  "stored in the .beam attribute, and the "
-                                  "units are set to Jy")
-                except:
-                    warnings.warn("Could not parse JY/BEAM unit.  Either you "
-                                  "should install the radio_beam package "
-                                  "or manually replace the units.  For now, "
-                                  "the units are being interpreted as Jy.")
-            else:
-                try:
-                    self._unit = u.Unit(self._meta['BUNIT'])
-                except ValueError:
-                    warnings.warn("Could not parse unit {0}".format(self._meta['BUNIT']))
-                    self._unit = None
-        elif hasattr(data, 'unit'):
-            self._unit = data.unit
+        # data must not be a quantity when stored in self._data
+        if hasattr(data, 'unit'):
             # strip the unit so that it can be treated as cube metadata
             data = data.value
-        else:
-            self._unit = None
 
         # TODO: mask should be oriented? Or should we assume correctly oriented here?
         self._data, self._wcs = cube_utils._orient(data, wcs)
@@ -158,6 +139,46 @@ class SpectralCube(object):
         if not isinstance(self._header, Header):
             raise TypeError("If a header is given, it must be a fits.Header")
 
+        # Beam loading must happen *after* WCS is read
+        if read_beam:
+            self._try_load_beam(header)
+
+        if 'BUNIT' in self._meta:
+
+            # special case: CASA (sometimes) makes non-FITS-compliant jy/beam headers
+            bunit = re.sub("\s", "", self._meta['BUNIT'].lower())
+            if bunit == 'jy/beam':
+                self._unit = u.Jy
+
+                if not read_beam:
+
+                    warnings.warn("Units are in Jy/beam. Attempting to parse "
+                                  "header for beam information.")
+
+                    self._try_load_beam(header)
+
+                    if hasattr(self, 'beam'):
+                        warnings.warn("Units were Jy/beam.  The 'beam' is now "
+                                      "stored in the .beam attribute, and the "
+                                      "units are set to Jy")
+                    else:
+                        warnings.warn("Could not parse Jy/beam unit.  Either "
+                                      "you should install the radio_beam "
+                                      "package or manually replace the units."
+                                      " For now, the units are being interpreted "
+                                      "as Jy.")
+
+            else:
+                try:
+                    self._unit = u.Unit(self._meta['BUNIT'])
+                except ValueError:
+                    warnings.warn("Could not parse unit {0}".format(self._meta['BUNIT']))
+                    self._unit = None
+        elif hasattr(data, 'unit'):
+            self._unit = data.unit
+        else:
+            self._unit = None
+
         # We don't pass the spectral unit via the initializer since the user
         # should be using ``with_spectral_unit`` if they want to set it.
         # However, we do want to keep track of what units the spectral axis
@@ -177,8 +198,7 @@ class SpectralCube(object):
 
     @property
     def _is_huge(self):
-        # TODO: make threshold depend on memory?
-        return cube_utils.is_huge(self, threshold=1e8)
+        return cube_utils.is_huge(self)
 
     def _new_cube_with(self, data=None, wcs=None, mask=None, meta=None,
                        fill_value=None, spectral_unit=None, unit=None):
@@ -224,6 +244,24 @@ class SpectralCube(object):
 
         return cube
 
+    def _try_load_beam(self, header):
+
+        try:
+            from radio_beam import Beam
+        except ImportError:
+            warnings.warn("radio_beam is not installed. No beam "
+                          "can be created.")
+
+        try:
+            self.beam = Beam.from_fits_header(header)
+            self._meta['beam'] = self.beam
+            self.pixels_per_beam = (self.beam.sr /
+                                    (wcs.utils.proj_plane_pixel_area(self.wcs) *
+                                     u.deg**2)).to(u.dimensionless_unscaled).value
+        except Exception as ex:
+            warnings.warn("Could not parse beam information from header."
+                          "  Exception was: {0}".format(ex.__repr__()))
+
     @property
     def unit(self):
         """ The flux unit """
@@ -318,7 +356,9 @@ class SpectralCube(object):
                              reduce=True, how='auto',
                              projection=False,
                              unit=None,
-                             check_endian=False, **kwargs):
+                             check_endian=False,
+                             progressbar=False,
+                             **kwargs):
         """
         Apply a numpy function to the cube
 
@@ -352,6 +392,9 @@ class SpectralCube(object):
             A flag to check the endianness of the data before applying the
             function.  This is only needed for optimized functions, e.g. those
             in the `bottleneck` package.
+        progressbar : bool
+            Show a progressbar while iterating over the slices through the
+            cube?
         kwargs : dict
             Passed to the numpy function.
 
@@ -379,9 +422,8 @@ class SpectralCube(object):
 
         if strategy == 'slice' and reduce:
             try:
-                out = self._reduce_slicewise(function, fill,
-                                              check_endian,
-                                              **kwargs)
+                out = self._reduce_slicewise(function, fill, check_endian,
+                                             progressbar=progressbar, **kwargs)
             except NotImplementedError:
                 pass
         elif how not in ['auto', 'cube']:
@@ -400,6 +442,7 @@ class SpectralCube(object):
                 return out
         elif projection and reduce:
             meta = {'collapse_axis': axis}
+            meta.update(self._meta)
 
             if hasattr(axis, '__len__') and len(axis) == 2:
                 # if operation is over two spatial dims
@@ -426,7 +469,7 @@ class SpectralCube(object):
 
 
     def _reduce_slicewise(self, function, fill, check_endian,
-                          includemask=False, **kwargs):
+                          includemask=False, progressbar=False, **kwargs):
         """
         Compute a numpy aggregation by grabbing one slice at a time
         """
@@ -444,8 +487,17 @@ class SpectralCube(object):
         else:
             planes = self._iter_slices(ax, fill=fill, check_endian=check_endian)
         result = next(planes)
+
+        if progressbar:
+            progressbar = ProgressBar(self.shape[ax])
+            pbu = progressbar.update
+        else:
+            pbu = lambda: True
+
+
         for plane in planes:
             result = function(np.dstack((result, plane)), axis=2, **kwargs)
+            pbu()
 
         if full_reduce:
             result = function(result)
@@ -511,8 +563,10 @@ class SpectralCube(object):
             out = ttl / counts
             if projection:
                 new_wcs = wcs_utils.drop_axis(self._wcs, np2wcs[axis])
+                meta = {'collapse_axis': axis}
+                meta.update(self._meta)
                 return Projection(out, copy=False, wcs=new_wcs,
-                                  meta={'collapse_axis': axis},
+                                  meta=meta,
                                   unit=self.unit, header=self._nowcs_header)
             else:
                 return out
@@ -543,10 +597,10 @@ class SpectralCube(object):
 
         if how == 'slice':
             if axis is None:
-                 raise NotImplementedError("The overall standard deviation "
-                                           "cannot be computed in a slicewise "
-                                           "manner.  Please use a "
-                                           "different strategy.")
+                raise NotImplementedError("The overall standard deviation "
+                                          "cannot be computed in a slicewise "
+                                          "manner.  Please use a "
+                                          "different strategy.")
             counts = self._count_nonzero_slicewise(axis=axis)
             ttl = self.apply_numpy_function(np.nansum, fill=np.nan, how='slice',
                                             axis=axis, unit=None,
@@ -564,8 +618,10 @@ class SpectralCube(object):
 
             if projection:
                 new_wcs = wcs_utils.drop_axis(self._wcs, np2wcs[axis])
+                meta = {'collapse_axis': axis}
+                meta.update(self._meta)
                 return Projection(out, copy=False, wcs=new_wcs,
-                                  meta={'collapse_axis': axis},
+                                  meta=meta,
                                   unit=self.unit, header=self._nowcs_header)
             else:
                 return out
@@ -709,7 +765,7 @@ class SpectralCube(object):
         return self._new_cube_with(data=data, unit=unit)
 
     def apply_function(self, function, axis=None, weights=None, unit=None,
-                       projection=False, **kwargs):
+                       projection=False, progressbar=False, **kwargs):
         """
         Apply a function to valid data along the specified axis or to the whole
         cube, optionally using a weight array that is the same shape (or at
@@ -730,6 +786,9 @@ class SpectralCube(object):
             should return quantities with units.
         projection : bool
             Return a projection if the resulting array is 2D?
+        progressbar : bool
+            Show a progressbar while iterating over the slices/rays through the
+            cube?
 
         Returns
         -------
@@ -752,22 +811,30 @@ class SpectralCube(object):
         # allocate memory for output array
         out = np.empty([nx, ny]) * np.nan
 
+        if progressbar:
+            progressbar = ProgressBar(nx*ny)
+            pbu = progressbar.update
+        else:
+            pbu = lambda: True
+
         # iterate over "lines of sight" through the cube
-        for x, y, slc in self._iter_rays(axis):
+        for y, x, slc in self._iter_rays(axis):
             # acquire the flattened, valid data for the slice
             data = self.flattened(slc, weights=weights)
             if len(data) != 0:
                 result = function(data, **kwargs)
                 if hasattr(result, 'value'):
                     # store result in array
-                    out[x, y] = result.value
+                    out[y, x] = result.value
                 else:
-                    out[x, y] = result
+                    out[y, x] = result
+            pbu()
 
         if projection and axis in (0,1,2):
             new_wcs = wcs_utils.drop_axis(self._wcs, np2wcs[axis])
 
             meta = {'collapse_axis': axis}
+            meta.update(self._meta)
 
             return Projection(out, copy=False, wcs=new_wcs, meta=meta,
                               unit=unit, header=self._nowcs_header)
@@ -779,15 +846,15 @@ class SpectralCube(object):
         Iterate over view corresponding to lines-of-sight through a cube
         along the specified axis
         """
-        nx, ny = self._get_flat_shape(axis)
+        ny, nx = self._get_flat_shape(axis)
 
-        for x in xrange(nx):
-            for y in xrange(ny):
+        for y in xrange(ny):
+            for x in xrange(nx):
                 # create length-1 view for each position
-                slc = [slice(x, x + 1), slice(y, y + 1)]
+                slc = [slice(y, y + 1), slice(x, x + 1), ]
                 # create a length-N slice (all-inclusive) along the selected axis
                 slc.insert(axis, slice(None))
-                yield x, y, slc
+                yield y, x, slc
 
     def _iter_slices(self, axis, fill=np.nan, check_endian=False):
         """
@@ -847,7 +914,7 @@ class SpectralCube(object):
                                      view=slice)
         return lat,lon,spec
 
-    def median(self, axis=None, **kwargs):
+    def median(self, axis=None, iterate_rays=False, **kwargs):
         """
         Compute the median of an array, optionally along an axis.
 
@@ -857,6 +924,9 @@ class SpectralCube(object):
         ----------
         axis : int (optional)
             The axis to collapse
+        iterate_rays : bool
+            Iterate over individual rays?  This mode is slower but can save RAM
+            costs, which may be extreme for large cubes
 
         Returns
         -------
@@ -865,15 +935,30 @@ class SpectralCube(object):
         """
         try:
             from bottleneck import nanmedian
-            result = self.apply_numpy_function(nanmedian, axis=axis,
-                                               projection=True,
-                                               check_endian=True, **kwargs)
+            bnok = True
         except ImportError:
-            result = self.apply_function(np.median, axis=axis, **kwargs)
+            bnok = False
+
+        # slicewise median is nonsense, must force how = 'cube'
+        if bnok and not iterate_rays:
+            log.debug("Using bottleneck nanmedian")
+            result = self.apply_numpy_function(nanmedian, axis=axis,
+                                               projection=True, unit=self.unit,
+                                               how='cube', check_endian=True,
+                                               **kwargs)
+        elif hasattr(np, 'nanmedian') and not iterate_rays:
+            log.debug("Using numpy nanmedian")
+            result = self.apply_numpy_function(np.nanmedian, axis=axis,
+                                               projection=True, unit=self.unit,
+                                               how='cube',**kwargs)
+        else:
+            log.debug("Using numpy median iterating over rays")
+            result = self.apply_function(np.median, projection=True, axis=axis,
+                                         unit=self.unit, **kwargs)
 
         return result
 
-    def percentile(self, q, axis=None, **kwargs):
+    def percentile(self, q, axis=None, iterate_rays=False, **kwargs):
         """
         Return percentiles of the data.
 
@@ -883,8 +968,21 @@ class SpectralCube(object):
             The percentile to compute
         axis : int, or None
             Which axis to compute percentiles over
+        iterate_rays : bool
+            Iterate over individual rays?  This mode is slower but can save RAM
+            costs, which may be extreme for large cubes
         """
-        return self.apply_function(np.percentile, q=q, axis=axis, **kwargs)
+        if hasattr(np, 'nanpercentile') and not iterate_rays:
+            result = self.apply_numpy_function(np.nanpercentile, q=q,
+                                               axis=axis, projection=True,
+                                               unit=self.unit, how='cube',
+                                               **kwargs)
+        else:
+            result = self.apply_function(np.percentile, q=q, axis=axis,
+                                         projection=True, unit=self.unit,
+                                         **kwargs)
+
+        return result
 
     def with_mask(self, mask, inherit_mask=True):
         """
@@ -953,7 +1051,7 @@ class SpectralCube(object):
                                     copy=False,
                                     unit=self.unit,
                                     meta=meta)
-                                
+
             # only one element, so drop an axis
             newwcs = wcs_utils.drop_axis(self._wcs, intslices[0])
             header = self._nowcs_header
@@ -1039,6 +1137,10 @@ class SpectralCube(object):
         from .spectral_axis import (convert_spectral_axis,
                                     determine_ctype_from_vconv)
 
+        # Allow string specification of units, for example
+        if not isinstance(unit, u.Unit):
+            unit = u.Unit(unit)
+
         # Velocity conventions: required for frq <-> velo
         # convert_spectral_axis will handle the case of no velocity
         # convention specified & one is required
@@ -1072,10 +1174,13 @@ class SpectralCube(object):
         newwcs = convert_spectral_axis(self._wcs, unit, out_ctype,
                                        rest_value=rest_value)
 
-        newmask = self._mask.with_spectral_unit(unit,
-                                                velocity_convention=vc,
-                                                rest_value=rest_value)
-        newmask._wcs = newwcs
+        if self._mask is not None:
+            newmask = self._mask.with_spectral_unit(unit,
+                                                    velocity_convention=vc,
+                                                    rest_value=rest_value)
+            newmask._wcs = newwcs
+        else:
+            newmask = None
 
         newwcs.wcs.set()
         cube = self._new_cube_with(wcs=newwcs, mask=newmask, meta=meta,
@@ -1147,6 +1252,7 @@ class SpectralCube(object):
         # Start off by extracting the world coordinates of the pixels
         _, lat, lon = self.world[0, :, :]
         spectral, _, _ = self.world[:, 0, 0]
+        spectral -= spectral[0] # offset from first pixel
 
         # Convert to radians
         lon = np.radians(lon)
@@ -1165,13 +1271,37 @@ class SpectralCube(object):
         x[:, 1:] = np.cumsum(np.degrees(dx), axis=1)
         y[1:, :] = np.cumsum(np.degrees(dy), axis=0)
 
-        x = x.reshape(1, x.shape[0], x.shape[1])
-        y = y.reshape(1, y.shape[0], y.shape[1])
-        spectral = spectral.reshape(-1, 1, 1) - spectral.ravel()[0]
-        x, y, spectral = np.broadcast_arrays(x, y, spectral)
+        x, y, spectral = np.broadcast_arrays(x[None,:,:], y[None,:,:], spectral[:,None,None])
 
         return spectral, y, x
 
+    @cached
+    def _pix_size_slice(self, axis):
+        """
+        Return the size of each pixel along any given direction.  Assumes
+        pixels have equal size.  Also assumes that the spectral and spatial
+        directions are separable, which is enforced throughout this code.
+
+        Parameters
+        ----------
+        axis : 0, 1, or 2
+            The axis along which to compute the pixel size
+
+        Returns
+        -------
+        Pixel size in units of either degrees or the appropriate spectral unit
+        """
+        if axis == 0:
+            # note that self._spectral_scale is required here because wcs
+            # forces into units of m, m/s, or Hz
+            return np.abs(self.wcs.pixel_scale_matrix[2,2]) * self._spectral_scale
+        elif axis in (1,2):
+            # the pixel size is a projection.  I think the pixel_scale_matrix
+            # must be symmetric, such that psm[axis,:]**2 == psm[:,axis]**2
+            return np.sum(self.wcs.pixel_scale_matrix[2-axis,:]**2)**0.5
+        else:
+            raise ValueError("Cubes have 3 axes.")
+
     @cached
     def _pix_size(self):
         """
@@ -1237,16 +1367,16 @@ class SpectralCube(object):
 
         _, _, spectral = self._wcs.all_pix2world(xpix, ypix, zpix, 0)
 
-        dspectral = np.diff(spectral)
+        # Take spectral units into account
+        # order of operations here is crucial!  If this is done after
+        # broadcasting, the full array size is allocated, which is bad!
+        dspectral = np.diff(spectral) * self._spectral_scale
 
         dx = np.abs(np.degrees(dx.reshape(1, dx.shape[0], dx.shape[1])))
         dy = np.abs(np.degrees(dy.reshape(1, dy.shape[0], dy.shape[1])))
         dspectral = np.abs(dspectral.reshape(-1, 1, 1))
         dx, dy, dspectral = np.broadcast_arrays(dx, dy, dspectral)
 
-        # Take spectral units into account
-        dspectral = dspectral * self._spectral_scale
-
         return dspectral, dy, dx
 
     def moment(self, order=0, axis=0, how='auto'):
@@ -1337,6 +1467,7 @@ class SpectralCube(object):
         meta = {'moment_order': order,
                 'moment_axis': axis,
                 'moment_method': how}
+        meta.update(self._meta)
 
         return Projection(out, copy=False, wcs=new_wcs, meta=meta,
                           header=self._nowcs_header)
@@ -1841,11 +1972,14 @@ class SpectralCube(object):
             :func:`~astropy.io.fits.open`.
         """
         from .io.core import read
+        from .stokes_spectral_cube import StokesSpectralCube
         cube = read(filename, format=format, hdu=hdu, **kwargs)
         if isinstance(cube, StokesSpectralCube):
-            return SpectralCube(data=cube._data, wcs=cube._wcs,
-                                meta=cube._meta, mask=cube._mask,
-                                header=cube._header)
+            if hasattr(cube, 'I'):
+                warnings.warn("Cube is a Stokes cube, returning spectral cube for I component")
+                return cube.I
+            else:
+                raise ValueError("Spectral cube is a Stokes cube that does not have an I component")
         else:
             return cube
 
@@ -1991,8 +2125,8 @@ class SpectralCube(object):
                     return self._glue_app
 
             glue_app.add_datasets(self._glue_app.data_collection, result)
-        
-        
+
+
     def to_pvextractor(self):
         """
         Open the cube in a quick viewer written in matplotlib that allows you
@@ -2017,7 +2151,10 @@ class SpectralCube(object):
         newframe: bool
             Send the cube to a new frame or to the current frame?
         """
-        import ds9
+        try:
+            import ds9
+        except ImportError:
+            import pyds9 as ds9
 
         if ds9id is None:
             dd = ds9.ds9(start=True)
@@ -2061,6 +2198,9 @@ class SpectralCube(object):
             header['CRVAL3'] *= self._spectral_scale
             header['CUNIT3'] = self._spectral_unit.to_string(format='FITS')
 
+        if 'beam' in self._meta:
+            header = self._meta['beam'].attach_to_header(header)
+
         # TODO: incorporate other relevant metadata here
         return header
 
@@ -2077,7 +2217,7 @@ class SpectralCube(object):
         Return the cube converted to the given unit (assuming it is equivalent).
         If conversion was required, this will be a copy, otherwise it will
         """
-        
+
         if not isinstance(unit, u.Unit):
             unit = u.Unit(unit)
 
@@ -2136,67 +2276,6 @@ class SpectralCube(object):
         return result
 
 
-class StokesSpectralCube(SpectralCube):
-
-    """
-    A class to store a spectral cube with multiple Stokes parameters. By
-    default, this will act like a Stokes I spectral cube, but other stokes
-    parameters can be accessed with attribute notation.
-    """
-
-    def __init__(self, data, wcs, mask=None, meta=None, header=None):
-
-        # WCS should be 3-d, data should be dict of 3-d, mask should be disk
-        # of 3-d
-
-        # XXX: For now, let's just extract I and work with that
-
-        super(StokesSpectralCube, self).__init__(data["I"], wcs,
-                                                 mask=mask["I"], meta=meta,
-                                                 header=header)
-
-        # TODO: deal with the other stokes parameters here
-
-    @classmethod
-    def read(cls, filename, format=None, hdu=None):
-        """
-        Read a spectral cube from a file.
-
-        If the file contains Stokes axes, they will be read in. If you are
-        only interested in the unpolarized emission (I), you can use
-        :meth:`~spectral_cube.SpectralCube.read` instead.
-
-        Parameters
-        ----------
-        filename : str
-            The file to read the cube from
-        format : str
-            The format of the file to read. (Currently limited to 'fits' and 'casa_image')
-        hdu : int or str
-            For FITS files, the HDU to read in (can be the ID or name of an
-            HDU).
-
-        Returns
-        -------
-        cube : :class:`SpectralCube`
-        """
-        raise NotImplementedError("")
-
-    def write(self, filename, overwrite=False, format=None):
-        """
-        Write the spectral cube to a file.
-
-        Parameters
-        ----------
-        filename : str
-            The path to write the file to
-        format : str
-            The format of the file to write. (Currently limited to 'fits')
-        overwrite : bool
-            If True, overwrite `filename` if it exists
-        """
-        raise NotImplementedError("")
-
 def determine_format_from_filename(filename):
     if filename[-4:] == 'fits':
         return 'fits'
