@@ -1,12 +1,15 @@
-import pytest
+from __future__ import print_function, absolute_import, division
+
 import operator
 import itertools
 import warnings
 import mmap
+from distutils.version import StrictVersion, LooseVersion
 
 # needed to test for warnings later
 warnings.simplefilter('always', UserWarning)
 
+import pytest
 
 from astropy.io import fits
 from astropy import units as u
@@ -23,31 +26,31 @@ from .. import spectral_axis
 from . import path
 from .helpers import assert_allclose, assert_array_equal
 
-from distutils.version import StrictVersion
 
 try:
     import yt
-    ytOK = True
-    yt_version = (StrictVersion(yt.__version__[:-4])
-                  if 'dev' in yt.__version__ else
-                  StrictVersion(yt.__version__))
+    YT_INSTALLED = True
+    YT_LT_301 = LooseVersion(yt.__version__) < LooseVersion('3.0.1')
 except ImportError:
-    yt_version = StrictVersion('0.0.0')
-    ytOK = False
+    YT_INSTALLED = False
+    YT_LT_301 = False
 
 try:
     import bottleneck
-    bottleneckOK = True
+    BOTTLENECK_INSTALLED = True
 except ImportError:
-    bottleneckOK = False
+    BOTTLENECK_INSTALLED = False
 
 try:
-    import radio_beam
+    from radio_beam import Beam
     RADIO_BEAM_INSTALLED = True
 except ImportError:
     RADIO_BEAM_INSTALLED = False
 
 
+NUMPY_LT_19 = LooseVersion(np.__version__) < LooseVersion('1.9.0')
+
+
 def cube_and_raw(filename):
     p = path(filename)
 
@@ -74,18 +77,31 @@ def test_huge_disallowed():
 
     cube, data = cube_and_raw('vda_Jybeam_lower.fits')
 
-    data = np.empty([1e2,1e3,1e3])
     cube = SpectralCube(data=data, wcs=cube.wcs)
 
-    assert cube._is_huge
+    assert not cube._is_huge
 
-    with pytest.raises(ValueError) as exc:
+    # We need to reduce the memory threshold rather than use a large cube to
+    # make sure we don't use too much memory during testing.
+    from .. import cube_utils
+    OLD_MEMORY_THRESHOLD = cube_utils.MEMORY_THRESHOLD
+
+    try:
+        cube_utils.MEMORY_THRESHOLD = 10
+
+        assert cube._is_huge
+
+        with pytest.raises(ValueError) as exc:
+            cube + 5*cube.unit
+        assert 'entire cube into memory' in exc.value.args[0]
+
+        cube.allow_huge_operations = True
+
+        # just make sure it doesn't fail
         cube + 5*cube.unit
-    assert 'entire cube into memory' in exc.value.args[0]
+    finally:
+        cube_utils.MEMORY_THRESHOLD = OLD_MEMORY_THRESHOLD
 
-    cube.allow_huge_operations = True
-    # just make sure it doesn't fail
-    cube + 5*cube.unit
 
 class BaseTest(object):
 
@@ -161,13 +177,15 @@ class TestSpectralCube(object):
             assert_allclose(w1, w2)
 
 
-    @pytest.mark.parametrize(('name','masktype'),
+    @pytest.mark.parametrize(('name','masktype','unit'),
                              itertools.product(('advs', 'dvsa', 'sdav', 'sadv', 'vsad', 'vad', 'adv',),
-                                               (BooleanArrayMask, LazyMask, FunctionMask, CompositeMask))
+                                               (BooleanArrayMask, LazyMask, FunctionMask, CompositeMask),
+                                               ('Hz', u.Hz),
+                                              )
                             )
-    def test_with_spectral_unit(self, name, masktype):
+    def test_with_spectral_unit(self, name, masktype, unit):
         cube, data = cube_and_raw(name + '.fits')
-        cube_freq = cube.with_spectral_unit(u.Hz)
+        cube_freq = cube.with_spectral_unit(unit)
 
         if masktype == BooleanArrayMask:
             mask = BooleanArrayMask(data>0, wcs=cube._wcs)
@@ -181,7 +199,7 @@ class TestSpectralCube(object):
             mask = CompositeMask(mask1, mask2)
 
         cube2 = cube.with_mask(mask)
-        cube_masked_freq = cube2.with_spectral_unit(u.Hz)
+        cube_masked_freq = cube2.with_spectral_unit(unit)
 
         assert cube_freq._wcs.wcs.ctype[cube_freq._wcs.wcs.spec] == 'FREQ-W2F'
         assert cube_masked_freq._wcs.wcs.ctype[cube_masked_freq._wcs.wcs.spec] == 'FREQ-W2F'
@@ -300,6 +318,9 @@ class TestArithmetic(object):
         assert np.all(d2 == c2.filled_data[:].value)
         assert c2.unit == u.K
 
+        # regression test #251: the _data attribute must not be a quantity
+        assert not hasattr(c2._data, 'unit')
+
     def test_subtract_cubes(self):
         d2 = self.d1 - self.d1
         c2 = self.c1 - self.c1
@@ -307,6 +328,9 @@ class TestArithmetic(object):
         assert np.all(c2.filled_data[:].value == 0)
         assert c2.unit == u.K
 
+        # regression test #251: the _data attribute must not be a quantity
+        assert not hasattr(c2._data, 'unit')
+
     @pytest.mark.parametrize(('value'),(1,1.0,2,2.0))
     def test_mul(self, value):
         d2 = self.d1 * value
@@ -379,10 +403,8 @@ class TestFilters(BaseTest):
         expected = d[d > 0.5] ** 2
         assert_allclose(c.flattened(weights=d), expected)
 
-    @pytest.mark.xfail
     def test_slice(self):
         c, d = self.c, self.d
-
         expected = d[:3, :2, ::2]
         expected = expected[expected > 0.5]
         assert_allclose(c[0:3, 0:2, 0::2].flattened(), expected)
@@ -421,23 +443,67 @@ class TestNumpyMethods(BaseTest):
         d = np.where(self.d > 0.5, self.d, 10)
         self._check_numpy(self.c.argmin, d, np.nanargmin)
 
-    def test_median(self):
-        m = np.empty(self.d.sum(axis=0).shape)
+    @pytest.mark.parametrize('iterate_rays', (True,False))
+    def test_median(self, iterate_rays):
+        # Make sure that medians ignore empty/bad/NaN values
+        m = np.empty(self.d.shape[1:])
         for y in range(m.shape[0]):
             for x in range(m.shape[1]):
                 ray = self.d[:, y, x]
+                # the cube mask is for values >0.5
                 ray = ray[ray > 0.5]
                 m[y, x] = np.median(ray)
-        assert_allclose(self.c.median(axis=0), m)
+        scmed = self.c.median(axis=0, iterate_rays=iterate_rays)
+        assert_allclose(scmed, m)
+        assert not np.any(np.isnan(scmed.value))
+        assert scmed.unit == self.c.unit
 
-    def test_percentile(self):
+    @pytest.mark.skipif('NUMPY_LT_19')
+    def test_bad_median_apply(self):
+        # this is a test for manually-applied numpy medians, which are different
+        # from the cube.median method that does "the right thing"
+        #
+        # for regular median, we expect a failure, which is why we don't use
+        # regular median.
+
+        scmed = self.c.apply_numpy_function(np.median, axis=0)
+        if StrictVersion(np.__version__) <= StrictVersion('1.9.3'):
+            assert np.count_nonzero(np.isnan(scmed)) == 5
+        else:
+            assert np.count_nonzero(np.isnan(scmed)) == 6
+
+        scmed = self.c.apply_numpy_function(np.nanmedian, axis=0)
+        assert np.count_nonzero(np.isnan(scmed)) == 0
+
+        # use a more aggressive mask to force there to be some all-nan axes
+        m2 = self.c>0.65*self.c.unit
+        scmed = self.c.with_mask(m2).apply_numpy_function(np.nanmedian, axis=0)
+        assert np.count_nonzero(np.isnan(scmed)) == 1
+
+    @pytest.mark.parametrize('iterate_rays', (True,False))
+    def test_bad_median(self, iterate_rays):
+        # This should have the same result as np.nanmedian, though it might be
+        # faster if bottleneck loads
+        scmed = self.c.median(axis=0, iterate_rays=iterate_rays)
+        assert np.count_nonzero(np.isnan(scmed)) == 0
+
+        m2 = self.c>0.65*self.c.unit
+        scmed = self.c.with_mask(m2).median(axis=0, iterate_rays=iterate_rays)
+        assert np.count_nonzero(np.isnan(scmed)) == 1
+
+    @pytest.mark.parametrize(('pct', 'iterate_rays'),
+                             (zip((3,25,50,75,97)*2,(True,)*5 + (False,)*5)))
+    def test_percentile(self, pct, iterate_rays):
         m = np.empty(self.d.sum(axis=0).shape)
         for y in range(m.shape[0]):
             for x in range(m.shape[1]):
                 ray = self.d[:, y, x]
                 ray = ray[ray > 0.5]
-                m[y, x] = np.percentile(ray, 3)
-        assert_allclose(self.c.percentile(3, axis=0), m)
+                m[y, x] = np.percentile(ray, pct)
+        scpct = self.c.percentile(pct, axis=0, iterate_rays=iterate_rays)
+        assert_allclose(scpct, m)
+        assert not np.any(np.isnan(scpct.value))
+        assert scpct.unit == self.c.unit
 
     @pytest.mark.parametrize('method', ('sum', 'min', 'max',
                              'median', 'argmin', 'argmax'))
@@ -512,7 +578,7 @@ SpectralCube with shape=(4, 3, 2) and unit=Jy:
         """.strip()
 
 
-@pytest.mark.skipif(not ytOK, reason='Could not import yt')
+@pytest.mark.skipif('not YT_INSTALLED')
 class TestYt():
     def setup_method(self, method):
         self.cube = SpectralCube.read(path('adv.fits'))
@@ -532,10 +598,11 @@ class TestYt():
         ds1,ds2,ds3 = ytc1.dataset, ytc2.dataset, ytc3.dataset
         assert_array_equal(ds1.domain_dimensions, ds2.domain_dimensions)
         assert_array_equal(ds2.domain_dimensions, ds3.domain_dimensions)
-        assert_allclose(ds1.domain_left_edge, ds2.domain_left_edge)
-        assert_allclose(ds2.domain_left_edge, ds3.domain_left_edge)
-        assert_allclose(ds1.domain_width, ds2.domain_width*np.array([1,1,1.0/self.spectral_factor]))
-        assert_allclose(ds1.domain_width, ds3.domain_width)
+        assert_allclose(ds1.domain_left_edge.value, ds2.domain_left_edge.value)
+        assert_allclose(ds2.domain_left_edge.value, ds3.domain_left_edge.value)
+        assert_allclose(ds1.domain_width.value,
+                        ds2.domain_width.value*np.array([1,1,1.0/self.spectral_factor]))
+        assert_allclose(ds1.domain_width.value, ds3.domain_width.value)
         assert self.nprocs == len(ds3.index.grids)
         assert ds1.spec_cube
         assert ds2.spec_cube
@@ -550,7 +617,7 @@ class TestYt():
         ds2.quan(1.0,unit2)
         ds3.quan(1.0,unit3)
 
-    @pytest.mark.skipif(yt_version < StrictVersion('3.0.1'), reason='yt 3.0 has a FITS-related bug')
+    @pytest.mark.skipif('YT_LT_301', reason='yt 3.0 has a FITS-related bug')
     def test_yt_fluxcompare(self):
         # Now check that we can compute quantities of the flux
         # and that they are equal
@@ -576,16 +643,15 @@ class TestYt():
         # Now test round-trip conversions between yt and world coordinates
         ytc1,ytc2,ytc3 = self.ytc1,self.ytc2,self.ytc3
         ds1,ds2,ds3 = ytc1.dataset, ytc2.dataset, ytc3.dataset
-        cube = self.cube
         yt_coord1 = ds1.domain_left_edge + np.random.random(size=3)*ds1.domain_width
         world_coord1 = ytc1.yt2world(yt_coord1)
-        assert_allclose(ytc1.world2yt(world_coord1), yt_coord1)
+        assert_allclose(ytc1.world2yt(world_coord1), yt_coord1.value)
         yt_coord2 = ds2.domain_left_edge + np.random.random(size=3)*ds2.domain_width
         world_coord2 = ytc2.yt2world(yt_coord2)
-        assert_allclose(ytc2.world2yt(world_coord2), yt_coord2)
+        assert_allclose(ytc2.world2yt(world_coord2), yt_coord2.value)
         yt_coord3 = ds3.domain_left_edge + np.random.random(size=3)*ds3.domain_width
         world_coord3 = ytc3.yt2world(yt_coord3)
-        assert_allclose(ytc3.world2yt(world_coord3), yt_coord3)
+        assert_allclose(ytc3.world2yt(world_coord3), yt_coord3.value)
 
 def test_read_write_rountrip(tmpdir):
     cube = SpectralCube.read(path('adv.fits'))
@@ -595,12 +661,15 @@ def test_read_write_rountrip(tmpdir):
 
     assert cube.shape == cube.shape
     assert_allclose(cube._data, cube2._data)
-    if ((hasattr(_wcs, '__version__') and StrictVersion(_wcs.__version__) != StrictVersion('5.9'))
-        or not hasattr(_wcs, '__version__')):
+    if (((hasattr(_wcs, '__version__')
+          and StrictVersion(_wcs.__version__) < StrictVersion('5.9'))
+         or not hasattr(_wcs, '__version__'))):
         # see https://github.com/astropy/astropy/pull/3992 for reasons:
         # we should upgrade this for 5.10 when the absolute accuracy is
         # maximized
         assert cube._wcs.to_header_string() == cube2._wcs.to_header_string()
+        # in 5.11 and maybe even 5.12, the round trip fails.  Maybe
+        # https://github.com/astropy/astropy/issues/4292 will solve it?
 
 @pytest.mark.parametrize(('memmap', 'base'),
                          ((True, mmap.mmap),
@@ -695,7 +764,7 @@ def test_preserve_spectral_unit():
     assert new_cube.spectral_axis.unit is u.GHz
 
 
-@pytest.mark.skipif(not bottleneckOK, reason='Bottleneck could not be imported')
+@pytest.mark.skipif('not BOTTLENECK_INSTALLED')
 def test_endians():
     """
     Test that the endianness checking returns something in Native form
@@ -713,7 +782,7 @@ def test_endians():
     mywcs.wcs.ctype[0] = 'RA'
     mywcs.wcs.ctype[1] = 'DEC'
     mywcs.wcs.ctype[2] = 'VELO'
-    
+
     bigcube = SpectralCube(data=big, wcs=mywcs)
     xbig = bigcube._get_filled_data(check_endian=True)
 
@@ -748,7 +817,7 @@ def test_slicing():
 
     sl = cube[:,1,:]
     assert sl.shape == (2,4)
-    
+
     v = cube[1:2,:,:]
     assert v.shape == (1,3,4)
 
@@ -759,11 +828,11 @@ def test_slicing():
 
 
 @pytest.mark.parametrize(('view','naxis'),
-                         [( (slice(None), 1, slice(None)), 2 ),
-                          ( (1, slice(None), slice(None)), 2 ),
-                          ( (slice(None), slice(None), 1), 2 ),
-                          ( (slice(None), slice(None), slice(1)), 3 ),
-                          ( (slice(1), slice(1), slice(1)), 3 ),
+                         [((slice(None), 1, slice(None)), 2),
+                          ((1, slice(None), slice(None)), 2),
+                          ((slice(None), slice(None), 1), 2),
+                          ((slice(None), slice(None), slice(1)), 3),
+                          ((slice(1), slice(1), slice(1)), 3),
                          ])
 def test_slice_wcs(view, naxis):
 
@@ -920,6 +989,26 @@ def test_preserve_bunit():
     assert cube.unit == u.Jy
     assert cube.header['BUNIT'] == 'Jy'
 
+@pytest.mark.skipif('not RADIO_BEAM_INSTALLED')
+def test_preserve_beam():
+
+    cube, data = cube_and_raw('advs.fits')
+
+    beam = Beam.from_fits_header(path("advs.fits"))
+
+    assert cube.beam == beam
+
+@pytest.mark.skipif('not RADIO_BEAM_INSTALLED')
+def test_append_beam_to_hdr():
+
+    cube, data = cube_and_raw('advs.fits')
+
+    orig_hdr = fits.getheader(path('advs.fits'))
+
+    assert cube.header['BMAJ'] == orig_hdr['BMAJ']
+    assert cube.header['BMIN'] == orig_hdr['BMIN']
+    assert cube.header['BPA'] == orig_hdr['BPA']
+
 def test_cube_with_swapped_axes():
     """
     Regression test for #208
@@ -948,3 +1037,76 @@ def test_jybeam_lower():
         assert hasattr(cube, 'beam')
         np.testing.assert_almost_equal(cube.beam.sr.value,
                                        (((1*u.arcsec/np.sqrt(8*np.log(2)))**2).to(u.sr)*2*np.pi).value)
+
+# Regression test for #257 (https://github.com/radio-astro-tools/spectral-cube/pull/257)
+def test_jybeam_whitespace():
+
+    cube, data = cube_and_raw('vda_Jybeam_whitespace.fits')
+
+    assert cube.unit == u.Jy
+    if RADIO_BEAM_INSTALLED:
+        assert hasattr(cube, 'beam')
+        np.testing.assert_almost_equal(cube.beam.sr.value,
+                                       (((1*u.arcsec/np.sqrt(8*np.log(2)))**2).to(u.sr)*2*np.pi).value)
+
+@pytest.mark.skipif('not RADIO_BEAM_INSTALLED')
+def test_beam_proj_meta():
+
+    cube, data = cube_and_raw('advs.fits')
+
+    moment = cube.moment0(axis=0)
+
+    # regression test for #250
+    assert 'beam' in moment.meta
+    assert 'BMAJ' in moment.hdu.header
+
+    slc = cube[0,:,:]
+
+    assert 'beam' in slc.meta
+
+    proj = cube.max(axis=0)
+
+    assert 'beam' in proj.meta
+
+def test_proj_meta():
+
+    cube, data = cube_and_raw('advs.fits')
+
+    moment = cube.moment0(axis=0)
+
+    assert 'BUNIT' in moment.meta
+    assert moment.meta['BUNIT'] == 'K'
+
+    slc = cube[0,:,:]
+
+    assert 'BUNIT' in slc.meta
+    assert slc.meta['BUNIT'] == 'K'
+
+    proj = cube.max(axis=0)
+
+    assert 'BUNIT' in proj.meta
+    assert proj.meta['BUNIT'] == 'K'
+
+def test_pix_sign():
+
+    cube, data = cube_and_raw('advs.fits')
+
+    s,y,x = (cube._pix_size_slice(ii) for ii in range(3))
+
+    assert s>0
+    assert y>0
+    assert x>0
+
+    cube.wcs.wcs.cdelt *= -1
+    s,y,x = (cube._pix_size_slice(ii) for ii in range(3))
+
+    assert s>0
+    assert y>0
+    assert x>0
+
+    cube.wcs.wcs.pc *= -1
+    s,y,x = (cube._pix_size_slice(ii) for ii in range(3))
+
+    assert s>0
+    assert y>0
+    assert x>0
